from __future__ import annotations
from pathlib import Path, PurePath, _PosixFlavour
from typing import Callable
import logging
import functools
import abc
from contextlib import contextmanager
import re
import os

logger = logging.getLogger(__name__)


class _URIFlavour(_PosixFlavour):
    has_drv = True
    is_supported = True

    def splitroot(self, part, sep=_PosixFlavour.sep):
        assert sep == self.sep

        match = re.match(rf"(.*):{re.escape(sep)}{{2}}(.*)", part)
        if match:
            drive, path = match.groups()
            drive = drive + "://"
            root = ""
            return drive, root, path
        else:
            return super().splitroot(part, sep=sep)


class URIBase(Path):
    _flavour = _URIFlavour()

    @property
    def scheme(self) -> str:
        if not self.drive:
            return ""
        return self.drive[:-len("://")]

    @property
    def path(self) -> str:
        begin = 1 if self.drive or self.root else 0
        return self.root + self._flavour.join(self.parts[begin:])

    def __repr__(self) -> str:
        s = ""
        if self.scheme:
            s += self.scheme + ":" + self._flavour.sep * 2
        s += self.path
        return "{}({!r})".format(self.__class__.__name__, s)


class PathManagerBase:

    def __init__(self):
        self._handlers = {}
        self._previous_path_managers = []

    def resolve(self, path: os.PathLike) -> Path:
        if not isinstance(path, URIBase):
            path = URIBase(path)
        if path.scheme:
            if path.scheme not in self._handlers:
                raise NotImplementedError
            return self._handlers[path.scheme](path)
        else:
            return Path(path.path)

    def register_handler(self, scheme: str) -> Callable:
        def decorator(func: Callable[[os.PathLike], Path]):
            self._handlers[scheme] = func
            logger.debug(f"Registered path handler for scheme {scheme}")
            return func
        return decorator

    def __enter__(self):
        self._previous_path_managers.append(PathManager._instance)
        PathManager._instance = self

    def __exit__(self, exc_type, exc_val, exc_tb):
        PathManager._instance = self._previous_path_managers.pop()


class PathManagerProxy(PathManagerBase):

    def __init__(self, instance: PathManagerBase):
        self._instance = instance

    def resolve(self, *args, **kwargs):
        return self._instance.resolve(*args, **kwargs)

    def register_handler(self, *args, **kwargs):
        return self._instance.register_handler(*args, **kwargs)

    def __enter__(self, *args, **kwargs):
        raise NotImplementedError

    def __exit__(self, *args, **kwargs):
        raise NotImplementedError


PathManager = PathManagerProxy(PathManagerBase())


class URI(URIBase):

    def _init(self, *args, **kwargs):
        self._local_path = PathManager.resolve(self)
        super()._init()

    def __str__(self) -> str:
        return str(self._local_path)


class PathTranslator(abc.ABC):

    def __init__(self, base_path: Path):
        self._base_path = base_path

    def __call__(self, url: URI) -> Path:
        return self._base_path / url.path


def register_translator(scheme: str, path: Path):
    class Translator(PathTranslator):
        def __init__(self):
            super().__init__(path)
    PathManager.register_handler(scheme)(Translator())



from __future__ import annotations
import os
from typing import Dict, Any
from pathlib import Path
from yacs.config import CfgNode as _CfgNode
import functools
import yaml

from .path_manager import URI

BASE_KEY = "_BASE_"


class CfgNode(_CfgNode):

    def __init__(self, init_dict=None, key_list=None, new_allowed=False):
        def ensure_cfgnode(item):
            if isinstance(item, dict) and not isinstance(item, _CfgNode):
                return CfgNode(item)
            else:
                return item
        if init_dict is not None:
            init_dict = {key: value if not isinstance(value, list) else list(map(ensure_cfgnode, value))
                         for key, value in init_dict.items()}

        super().__init__(init_dict=init_dict, key_list=key_list, new_allowed=new_allowed)

    @classmethod
    def load_yaml_with_base(cls, filename: os.PathLike) -> CfgNode:
        uri = URI(filename)
        with uri.open("r") as f:
            cfg = cls.load_cfg(f)
        if BASE_KEY in cfg:
            base_cfg = cls.load_yaml_with_base(cfg[BASE_KEY])
            del cfg[BASE_KEY]
            base_cfg.merge_from_other_cfg(cfg)
            return base_cfg
        return cfg

    def merge_with_dict(self, overrides: Dict[str, Any]):
        for key, value in overrides.items():
            child = self
            *path_segments, prop = key.split(".")
            for path_segment in path_segments:
                if path_segment not in child:
                    child[path_segment] = CfgNode()
                child = child[path_segment]
            child[prop] = value

    def dump(self, **kwargs):
        def convert_node(cfg_node):
            if isinstance(cfg_node, list):
                return [convert_node(v) for v in cfg_node]
            if isinstance(cfg_node, dict):
                return {k: convert_node(v) for k, v in cfg_node.items()}
            else:
                return cfg_node
        return yaml.safe_dump(convert_node(self), **kwargs)

    def params_dict(self):
        params = dict()
        for k, v in self.items():
            if isinstance(v, CfgNode):
                for child_k, child_v in v.params_dict().items():
                    params[k + "." + child_k] = child_v
            else:
                params[k] = v
        return params
